@@ -14,12 +14,12 @@
 use libipld::{Cid, DefaultParams};
 use rusqlite::{
     config::DbConfig,
-    params,
+    params, params_from_iter,
     types::FromSql,
     Connection,
     Error::{QueryReturnedNoRows, SqliteFailure},
     ErrorCode::DatabaseBusy,
-    OptionalExtension, ToSql, Transaction,
+    OptionalExtension, ToSql, Transaction, TransactionBehavior,
 };
 use std::{
     collections::{BTreeSet, HashSet},
@@ -32,10 +32,9 @@ use crate::{
     cache::{BlockInfo, CacheTracker},
     cidbytes::CidBytes,
     error::Context,
-    BlockStoreError, SizeTargets, StoreStats, Synchronous, TempPin,
+    BlockStoreError, SizeTargets, StoreStats, Synchronous,
 };
 use anyhow::Context as _;
-#[cfg(test)]
 use itertools::Itertools;
 
 const PRAGMAS: &str = r#"
@@ -164,8 +163,23 @@ fn get_id(txn: &Transaction, cid: impl ToSql) -> rusqlite::Result<Option<i64>> {
         .optional()
 }
 
+#[derive(Debug, PartialEq)]
+pub(crate) struct BlockStats {
+    count: u64,
+    size: u64,
+}
+
+impl From<StoreStats> for BlockStats {
+    fn from(s: StoreStats) -> Self {
+        Self {
+            count: s.count,
+            size: s.size,
+        }
+    }
+}
+
 /// returns the number and size of blocks, excluding orphaned blocks, computed from scratch
-pub(crate) fn compute_store_stats(txn: &Transaction) -> crate::Result<StoreStats> {
+pub(crate) fn compute_store_stats(txn: &Transaction) -> crate::Result<BlockStats> {
     let (count, size): (i64, i64) = txn
         .prepare(
             "SELECT COUNT(id), COALESCE(SUM(LENGTH(block)), 0) \
@@ -174,7 +188,7 @@ pub(crate) fn compute_store_stats(txn: &Transaction) -> crate::Result<StoreStats
         .ctx("computing store stats (prep)")?
         .query_row([], |row| Ok((row.get(0)?, row.get(1)?)))
         .ctx("computing store stats")?;
-    Ok(StoreStats {
+    Ok(BlockStats {
         count: u64::try_from(count).ctx("computed count")?,
         size: u64::try_from(size).ctx("computed size")?,
     })
@@ -184,7 +198,7 @@ pub(crate) fn compute_store_stats(txn: &Transaction) -> crate::Result<StoreStats
 pub(crate) fn recompute_store_stats(conn: &mut Connection) -> crate::Result<()> {
     let _span = tracing::debug_span!("check stats").entered();
     // first a read-only transaction to determine the true base
-    let (stats, truth) = in_txn(conn, None, |txn| {
+    let (stats, truth) = in_txn(conn, None, false, |txn| {
         let stats = get_store_stats(txn)?;
         let truth = compute_store_stats(txn)?;
         Ok((stats, truth))
@@ -192,9 +206,9 @@ pub(crate) fn recompute_store_stats(conn: &mut Connection) -> crate::Result<()>
 
     tracing::debug!("applying findings");
     // now compute the correction based on what the above snapshot has calculated
-    in_txn(conn, None, |txn| {
-        let stats2 = get_store_stats(txn)?;
-        let new_stats = StoreStats {
+    in_txn(conn, None, true, move |txn| {
+        let stats2 = BlockStats::from(get_store_stats(txn)?);
+        let new_stats = BlockStats {
             count: stats2.count - stats.count + truth.count,
             size: stats2.size - stats.size + truth.size,
         };
@@ -224,9 +238,21 @@ pub(crate) fn get_store_stats(txn: &Transaction) -> crate::Result<StoreStats> {
         .ctx("getting store stats (prep)")?
         .query_row([], |row| Ok((row.get(0)?, row.get(1)?)))
         .ctx("getting store stats")?;
+    let page_size = txn
+        .pragma_query_value(None, "page_size", |r| r.get::<_, i64>(0))
+        .ctx("getting page_size")?;
+    let used_pages = txn
+        .pragma_query_value(None, "page_count", |r| r.get::<_, i64>(0))
+        .ctx("getting page_count")?;
+    let free_pages = txn
+        .pragma_query_value(None, "freelist_count", |r| r.get::<_, i64>(0))
+        .ctx("getting freelist_count")?;
     let result = StoreStats {
         count: u64::try_from(count).ctx("getting count")?,
         size: u64::try_from(size).ctx("getting size")?,
+        page_size: u64::try_from(page_size).ctx("getting page_size")?,
+        used_pages: u64::try_from(used_pages).ctx("getting used_pages")?,
+        free_pages: u64::try_from(free_pages).ctx("getting free_pages")?,
     };
     Ok(result)
 }
@@ -256,7 +282,7 @@ pub(crate) fn incremental_gc(
 
     // get the store stats from the stats table:
     // if we don't exceed any of the size targets, there is nothing to do
-    let mut stats = in_txn(conn, None, get_store_stats)?;
+    let mut stats = in_txn(conn, None, false, get_store_stats)?;
     if !size_targets.exceeded(&stats) {
         tracing::info!(
             blocks = display(stats.count),
@@ -271,6 +297,7 @@ pub(crate) fn incremental_gc(
     let mut ids = in_txn(
         conn,
         Some(("getting unreferenced CIDs", Duration::from_secs(3))),
+        false,
         |txn| {
             // find all ids that are not pinned (directly or indirectly)
             let mut id_query = txn
@@ -305,7 +332,7 @@ pub(crate) fn incremental_gc(
 
     let mut n = 0;
     let mut ret_val = true;
-    for id in ids.iter() {
+    for id in ids.into_iter() {
         if n >= min_blocks && t0.elapsed() > max_duration {
             tracing::info!(removed = n, "stopping due to time constraint");
             ret_val = false;
@@ -315,10 +342,14 @@ pub(crate) fn incremental_gc(
             tracing::info!(removed = n, "finished, target reached");
             break;
         }
-        let res = in_txn(conn, None, |txn| {
-            // get block size and check whether now referenced
-            let mut block_size_stmt = c!("getting GC block (prep)" => txn.prepare_cached(
-                r#"
+        let res = in_txn(
+            conn,
+            Some(("", Duration::from_millis(100))),
+            true,
+            move |txn| {
+                // get block size and check whether now referenced
+                let mut block_size_stmt = c!("getting GC block (prep)" => txn.prepare_cached(
+                    r#"
                 WITH RECURSIVE
                     ancestor(id) AS (
                         SELECT ?
@@ -329,37 +360,38 @@ pub(crate) fn incremental_gc(
                 SELECT LENGTH(block), cid, (SELECT count(*) FROM names)
                     FROM cids, blocks ON id = block_id WHERE id = ?;
                 "#,
-            ));
-            let mut update_stats_stmt = c!("updating GC stats (prep)" =>
+                ));
+                let mut update_stats_stmt = c!("updating GC stats (prep)" =>
                 txn.prepare_cached("UPDATE stats SET count = count - 1, size = size - ?"));
-            let mut delete_stmt = c!("deleting GC block (prep)" => txn.prepare_cached("DELETE FROM blocks WHERE block_id = ?"));
+                let mut delete_stmt = c!("deleting GC block (prep)" => txn.prepare_cached("DELETE FROM blocks WHERE block_id = ?"));
 
-            tracing::trace!("deleting id {}", id);
+                tracing::trace!("deleting id {}", id);
 
-            let block_size: Option<(i64, CidBytes, i64)> = block_size_stmt
-                .query_row([id, id], |row| Ok((row.get(0)?, row.get(1)?, row.get(2)?)))
-                .optional()
-                .ctx("getting GC block")?;
-            tracing::trace!(block_size = ?&block_size);
-            if let Some((block_size, cid, names)) = block_size {
-                if names != 0 {
-                    // block is referenced again
-                    return Ok(None);
+                let block_size: Option<(i64, CidBytes, i64)> = block_size_stmt
+                    .query_row([id, id], |row| Ok((row.get(0)?, row.get(1)?, row.get(2)?)))
+                    .optional()
+                    .ctx("getting GC block")?;
+                tracing::trace!(block_size = ?&block_size);
+                if let Some((block_size, cid, names)) = block_size {
+                    if names != 0 {
+                        // block is referenced again
+                        return Ok(None);
+                    }
+                    let cid = Cid::try_from(&cid)?;
+                    let len = c!("getting GC block size" => usize::try_from(block_size));
+                    c!("updating GC stats" => update_stats_stmt.execute([block_size]));
+                    tracing::trace!("stats updated");
+                    c!("deleting GC block" => delete_stmt.execute(params![id]));
+                    Ok(Some((block_size, cid, len)))
+                } else {
+                    Ok(None)
                 }
-                let cid = Cid::try_from(&cid)?;
-                let len = c!("getting GC block size" => usize::try_from(block_size));
-                c!("updating GC stats" => update_stats_stmt.execute([block_size]));
-                tracing::trace!("stats updated");
-                c!("deleting GC block" => delete_stmt.execute(&[id]));
-                Ok(Some((block_size, cid, len)))
-            } else {
-                Ok(None)
-            }
-        })?;
+            },
+        )?;
         if let Some((size, cid, len)) = res {
             stats.count -= 1;
             stats.size -= size as u64;
-            cache_tracker.blocks_deleted(vec![BlockInfo::new(*id, &cid, len)]);
+            cache_tracker.blocks_deleted(vec![BlockInfo::new(id, &cid, len)]);
             n += 1;
         }
     }
@@ -367,23 +399,70 @@ pub(crate) fn incremental_gc(
     if n > 0 {
         // the above only removed the blocks, now we need to clean up those cids that we don’t
         // need anymore
-        in_txn(
+
+        // doing this in one transaction may block the DB for too long, so get the IDs first and then
+        // remove them in batches
+        let ids = in_txn(
             conn,
-            Some(("cleaning up CIDs", Duration::from_secs(1))),
+            Some(("getting IDs to clean up", Duration::from_secs(5))),
+            false,
             |txn| {
-                let mut del_cid = c!("deleting CIDs (prep)" => txn.prepare_cached(
-                    // refs.parent_id is not a blocker because if we delete this it means that
-                    // the block is gone
-                    "DELETE FROM cids WHERE \
+                let mut stmt = c!("getting IDs (prep)" => txn.prepare_cached(
+                        // refs.parent_id is not a blocker because if we delete this it means that
+                        // the block is gone
+                        "SELECT id FROM cids WHERE \
                         id NOT IN (SELECT block_id FROM blocks) AND \
                         id NOT IN (SELECT block_id FROM aliases) AND \
                         id NOT IN (SELECT child_id FROM refs) AND \
-                        id NOT IN (SELECT block_id FROM temp_pins)"
+                        id NOT IN (SELECT block_id FROM temp_pins)",
                 ));
-                c!("deleting CIDs" => del_cid.execute([]));
-                Ok(())
+                let ids = c!("getting IDs" => stmt.query_map([], |row| row.get(0)));
+                ids.collect::<Result<Vec<i64>, _>>().ctx("ids")
             },
         )?;
+
+        tracing::debug!("cleaning up {} IDs", ids.len());
+
+        // this number is linked to the prepared query below!
+        const BATCH_SIZE: usize = 10;
+        for ids in &ids.into_iter().chunks(BATCH_SIZE) {
+            let mut v = Vec::with_capacity(BATCH_SIZE);
+            v.extend(ids);
+            if v.len() == BATCH_SIZE {
+                in_txn(
+                    conn,
+                    Some(("cleaning up CIDs", Duration::from_millis(100))),
+                    false,
+                    move |txn| {
+                        let mut del_cid = c!("deleting CIDs (prep)" => txn.prepare_cached(
+                            "DELETE FROM cids WHERE \
+                                id in (VALUES (?), (?), (?), (?), (?), (?), (?), (?), (?), (?)) AND \
+                                id NOT IN (SELECT block_id FROM blocks) AND \
+                                id NOT IN (SELECT block_id FROM aliases) AND \
+                                id NOT IN (SELECT child_id FROM refs) AND \
+                                id NOT IN (SELECT block_id FROM temp_pins)"
+                        ));
+                        c!("deleting CIDs" => del_cid.execute(params_from_iter(v.iter())));
+                        Ok(())
+                    },
+                )?;
+            } else {
+                in_txn(conn, None, false, move |txn| {
+                    let mut stmt = c!("deleting CIDs (prep)" => txn.prepare_cached(
+                        "DELETE FROM cids WHERE \
+                            id = ? AND \
+                            id NOT IN (SELECT block_id FROM blocks) AND \
+                            id NOT IN (SELECT block_id FROM aliases) AND \
+                            id NOT IN (SELECT child_id FROM refs) AND \
+                            id NOT IN (SELECT block_id FROM temp_pins)"
+                    ));
+                    for id in v.iter() {
+                        c!("deleting CIDs" => stmt.execute([id]));
+                    }
+                    Ok(())
+                })?;
+            }
+        }
     }
 
     Ok(ret_val)
@@ -398,36 +477,36 @@ pub(crate) fn delete_temp_pin(txn: &Transaction, pin: i64) -> crate::Result<()>
 
 pub(crate) fn extend_temp_pin(
     txn: &Transaction,
-    pin: &mut TempPin,
+    mut id: i64,
     links: Vec<impl ToSql>,
-) -> crate::Result<()> {
+) -> crate::Result<i64> {
     for link in links {
         let block_id = c!("getting ID for temp pinning" => get_or_create_id(txn, link));
         // it is important that the above is a write action, because otherwise a rollback may
         // invalidate the id stored in the TempPin in the below
-        add_temp_pin(txn, block_id, pin).context("extending temp_pin")?;
+        id = add_temp_pin(txn, block_id, id).context("extending temp_pin")?;
     }
-    Ok(())
+    Ok(id)
 }
 
-fn add_temp_pin(txn: &Transaction, block_id: i64, pin: &mut TempPin) -> crate::Result<()> {
-    if pin.id > 0 {
+fn add_temp_pin(txn: &Transaction, block_id: i64, pin: i64) -> crate::Result<i64> {
+    if pin > 0 {
         txn.prepare_cached("INSERT OR IGNORE INTO temp_pins (id, block_id) VALUES (?, ?)")
             .ctx("extending existing temp_pin (prep)")?
-            .execute([pin.id, block_id])
+            .execute([pin, block_id])
             .ctx("extending existing temp_pin")?;
+        Ok(pin)
     } else {
         // we must not reuse IDs, but sqlite takes care of transactionality here
-        pin.id = txn
+        Ok(txn
             .prepare_cached(
                 "INSERT INTO temp_pins (id, block_id) VALUES \
                 ((SELECT coalesce(max(id), 0) FROM temp_pins) + 1, ?) RETURNING id",
             )
             .ctx("creating new temp_pin (prep)")?
             .query_row([block_id], |row| row.get(0))
-            .ctx("creating new temp_pin")?;
+            .ctx("creating new temp_pin")?)
     }
-    Ok(())
 }
 
 pub(crate) struct PutBlockResult {
@@ -442,8 +521,8 @@ pub(crate) fn put_block<C: ToSql>(
     key: &C,
     data: &[u8],
     links: impl IntoIterator<Item = C>,
-    pin: Option<&mut TempPin>,
-) -> crate::Result<PutBlockResult> {
+    mut pin: Option<i64>,
+) -> crate::Result<(Option<i64>, PutBlockResult)> {
     // this is important: we need write lock on the table so that add_temp_pin is never rolled back
     let block_id = c!("getting put_block ID" => get_or_create_id(txn, key));
     let block_exists = txn
@@ -475,15 +554,18 @@ pub(crate) fn put_block<C: ToSql>(
                 .ctx("adding put_block link")?;
         }
     }
-    if let Some(pin) = pin {
+    if let Some(pin) = pin.as_mut() {
         // create a temporary alias for the block, even if it already exists
         // this is only safe because get_or_create_id ensured that we have write lock on the table
-        add_temp_pin(txn, block_id, pin).context("adding put_block temp_pin")?;
+        *pin = add_temp_pin(txn, block_id, *pin).context("adding put_block temp_pin")?;
     }
-    Ok(PutBlockResult {
-        id: block_id,
-        block_exists,
-    })
+    Ok((
+        pin,
+        PutBlockResult {
+            id: block_id,
+            block_exists,
+        },
+    ))
 }
 
 /// Get a block
@@ -724,7 +806,13 @@ fn ws(s: impl AsRef<str>) -> String {
         }
         r.push_str(t);
     }
-    r
+    r.to_lowercase()
+        .replace("primary_key", "primary key")
+        // adding AUTOINCREMENT doesn’t really work, but it doesn’t hurt here either
+        // (only CREATE TABLE switched on AUTOINCREMENT special behaviour, but our CIDs
+        // don’t need monotonically incrementing IDs, unique is enough)
+        .replace("primary key autoincrement", "primary key")
+        .replace("unique", "primary key")
 }
 
 fn ensure_table(txn: &Transaction, name: &str, sql: &str) -> crate::Result<bool> {
@@ -742,7 +830,7 @@ fn ensure_table(txn: &Transaction, name: &str, sql: &str) -> crate::Result<bool>
             tracing::debug!("table {} is up-to-date", name);
             return Ok(false);
         }
-        if let Some(prefix) = ex_ws.find("CONSTRAINT") {
+        if let Some(prefix) = ex_ws.find("constraint") {
             // definitions must be equal up to the first constraint
             if ex_ws[..prefix] != sql_ws[..prefix] {
                 return Err(BlockStoreError::Other(anyhow::anyhow!(
@@ -756,7 +844,7 @@ fn ensure_table(txn: &Transaction, name: &str, sql: &str) -> crate::Result<bool>
             if sql_ws[..ex_trim.len()] != *ex_trim
                 || !sql_ws[ex_trim.len()..]
                     .trim_start_matches(|c| ", ".contains(c))
-                    .starts_with("CONSTRAINT")
+                    .starts_with("constraint")
             {
                 return Err(BlockStoreError::Other(anyhow::anyhow!(
                     "cannot update table `{}` due to incompatible data content",
@@ -784,17 +872,25 @@ fn ensure_tables(txn: &Transaction, tables: &[(&str, &str)]) -> crate::Result<()
     let mut changed = false;
 
     c!("writable schema" => txn.pragma_update(None, "writable_schema", true));
+    tracing::debug!("schema now read-write");
     for (name, sql) in tables {
         changed |=
             ensure_table(txn, *name, *sql).with_context(|| format!("ensuring table {}", name))?;
     }
+    tracing::debug!("schemas checked");
 
     if changed {
-        c!("increment schema version" => txn.pragma_update(None, "schema_version", version + 1));
+        let version = version + 1;
+        c!("increment schema version" => txn.pragma_update(None, "schema_version", version));
+        tracing::debug!("schema version updated to {}", version);
     }
     c!("writable schema" => txn.pragma_update(None, "writable_schema", false));
+    tracing::debug!("schema read-only again");
 
-    c!("integrity check" => txn.execute_batch("PRAGMA integrity_check"));
+    if changed {
+        c!("integrity check" => txn.execute_batch("PRAGMA integrity_check"));
+        tracing::debug!("integrity check done");
+    }
     Ok(())
 }
 
@@ -851,7 +947,7 @@ pub(crate) fn init_db(
 
     c!("foreign keys off" => conn.pragma_update(None, "foreign_keys", false));
 
-    in_txn(conn, Some(("init", Duration::from_secs(1))), |txn| {
+    in_txn(conn, Some(("init", Duration::from_secs(1))), true, |txn| {
         let user_version = c!("getting user_version" => user_version(txn));
         if user_version > 2 {
             return Err(anyhow::anyhow!(
@@ -890,7 +986,7 @@ pub(crate) fn init_db(
 
 pub(crate) fn integrity_check(conn: &mut Connection) -> crate::Result<Vec<String>> {
     let _span = tracing::debug_span!("db integrity check").entered();
-    in_txn(conn, None, |txn| {
+    in_txn(conn, None, false, |txn| {
         let mut stmt = c!("checking sqlite integrity (prep)" => txn.prepare("SELECT integrity_check FROM pragma_integrity_check"));
         let result = c!("checking sqlite integrity" => stmt.query_map([], |row| row.get(0)))
             .collect::<rusqlite::Result<Vec<String>>>()
@@ -907,12 +1003,17 @@ pub(crate) fn integrity_check(conn: &mut Connection) -> crate::Result<Vec<String
 ///
 /// this is an attempt to avoid spamming the log with lots of irrelevant info.
 /// execute a statement in a write transaction
-pub(crate) fn in_txn<T>(
+pub(crate) fn in_txn<T, F>(
     conn: &mut Connection,
     name: Option<(&str, Duration)>,
-    mut f: impl FnMut(&Transaction) -> crate::Result<T>,
-) -> crate::Result<T> {
-    let _span = if let Some((name, _)) = name {
+    immediate: bool,
+    f: F,
+) -> crate::Result<T>
+where
+    // since this function can be retried many times, it must no remember the Transaction nor reference other state
+    F: for<'a> Fn(&'a Transaction) -> crate::Result<T> + 'static,
+{
+    let _span = if let Some(name) = name.map(|x| x.0).filter(|x| !x.is_empty()) {
         tracing::debug_span!("txn", "{}", name).entered()
     } else {
         tracing::trace_span!("txn").entered()
@@ -920,7 +1021,9 @@ pub(crate) fn in_txn<T>(
     let started = Instant::now();
     let mut attempts = 0;
     loop {
-        let txn = c!("beginning transaction" => conn.transaction());
+        let txn = c!("beginning transaction" =>
+            if immediate { conn.transaction_with_behavior(TransactionBehavior::Immediate) } else { conn.transaction() }
+        );
         let result = f(&txn);
         let result = result.and_then(|t| {
             c!("committing transaction" => txn.commit());
@@ -949,7 +1052,7 @@ pub(crate) fn in_txn<T>(
                 }
             }
             Err(cause) => {
-                tracing::error!("transaction rolled back! {}", cause);
+                tracing::error!("transaction rolled back! {:#}", cause);
                 break Err(cause);
             }
         }
