@@ -67,6 +67,7 @@ use libipld::{codec::References, store::StoreParams, Block, Cid, Ipld};
 use parking_lot::Mutex;
 use rusqlite::{Connection, DatabaseName, OpenFlags};
 use std::{
+    borrow::Cow,
     collections::HashSet,
     fmt,
     iter::FromIterator,
@@ -74,7 +75,10 @@ use std::{
     mem,
     ops::DerefMut,
     path::{Path, PathBuf},
-    sync::Arc,
+    sync::{
+        atomic::{AtomicBool, Ordering},
+        Arc,
+    },
     time::Duration,
 };
 use tracing::*;
@@ -212,6 +216,7 @@ pub struct BlockStore<S> {
     expired_temp_pins: Arc<Mutex<Vec<i64>>>,
     config: Config,
     db_path: DbPath,
+    recompute_done: Arc<AtomicBool>,
     _s: PhantomData<S>,
 }
 
@@ -219,6 +224,9 @@ pub struct BlockStore<S> {
 pub struct StoreStats {
     count: u64,
     size: u64,
+    page_size: u64,
+    used_pages: u64,
+    free_pages: u64,
 }
 
 impl StoreStats {
@@ -231,6 +239,29 @@ impl StoreStats {
     pub fn size(&self) -> u64 {
         self.size
     }
+
+    /// Page size used by the SQLite DB
+    pub fn page_size(&self) -> u64 {
+        self.page_size
+    }
+
+    /// Number of used pages in the SQLite DB
+    ///
+    /// Multiply this with [`page_size`](#method.page_size) to obtain an upper bound
+    /// on how much space is actually used. The value returned by [`size`](#method.size)
+    /// will always be smaller than this, since it only counts net block data, without
+    /// overhead. A large difference suggests the need for calling `vacuum`.
+    pub fn used_pages(&self) -> u64 {
+        self.used_pages
+    }
+
+    /// Number of unused pages in the SQLite DB
+    ///
+    /// The DB file can be shrunk by at least this page count by calling `vacuum`, which often is
+    /// a long-running procedure.
+    pub fn free_pages(&self) -> u64 {
+        self.free_pages
+    }
 }
 
 /// a handle that contains a temporary pin
@@ -311,6 +342,7 @@ where
             expired_temp_pins: Arc::new(Mutex::new(Vec::new())),
             config,
             db_path,
+            recompute_done: Arc::new(AtomicBool::new(false)),
             _s: PhantomData,
         };
         if !is_memory {
@@ -319,12 +351,19 @@ where
                 if let Err(e) = recompute_store_stats(&mut conn.conn) {
                     tracing::error!("cannot recompute store stats: {}", e);
                 }
+                // This is done to avoid GC doing a wal_checkpoint(RESTART) while the above
+                // long-running query is ongoing, since that would block all writers during
+                // that period.
+                conn.recompute_done.store(true, Ordering::SeqCst);
             });
+        } else {
+            this.recompute_done.store(true, Ordering::SeqCst);
         }
         if this.config.cache_tracker.has_persistent_state() {
             let ids = in_txn(
                 &mut this.conn,
                 Some(("get IDs", Duration::from_secs(1))),
+                false,
                 get_ids,
             )?;
             this.config.cache_tracker.retain_ids(&ids);
@@ -356,6 +395,7 @@ where
             expired_temp_pins: self.expired_temp_pins.clone(),
             config: self.config.clone(),
             db_path: self.db_path.clone(),
+            recompute_done: self.recompute_done.clone(),
             _s: PhantomData,
         })
     }
@@ -400,6 +440,7 @@ where
         let ids = in_txn(
             &mut conn,
             Some(("get ids", Duration::from_secs(1))),
+            false,
             get_ids,
         )?;
         config.cache_tracker.retain_ids(&ids);
@@ -408,19 +449,20 @@ where
             expired_temp_pins: Arc::new(Mutex::new(Vec::new())),
             config,
             db_path: DbPath::Memory,
+            recompute_done: Arc::new(AtomicBool::new(true)),
             _s: PhantomData,
         })
     }
 
-    pub fn backup(&mut self, path: impl AsRef<Path>) -> Result<()> {
-        in_txn(&mut self.conn, None, move |txn| {
-            txn.backup(DatabaseName::Main, path.as_ref(), None)
+    pub fn backup(&mut self, path: PathBuf) -> Result<()> {
+        in_txn(&mut self.conn, None, false, move |txn| {
+            txn.backup(DatabaseName::Main, path.as_path(), None)
                 .ctx("backing up DB")
         })
     }
 
     pub fn flush(&mut self) -> crate::Result<()> {
-        in_txn(&mut self.conn, None, |txn| {
+        in_txn(&mut self.conn, None, false, |txn| {
             txn.pragma_update(None, "wal_checkpoint", &"TRUNCATE")
                 .ctx("flushing WAL")
         })
@@ -465,6 +507,7 @@ where
         in_txn(
             &mut self.conn,
             Some(("dropping expired temp_pins", Duration::from_millis(100))),
+            true,
             move |txn| {
                 // get rid of dropped temp aliases, this should be fast
                 for id in expired_temp_pins.iter() {
@@ -480,6 +523,7 @@ where
     /// This is the same as running incremental GC without limits, plus a full SQLITE VACUUM.
     pub fn gc(&mut self) -> Result<()> {
         self.cleanup_temp_pins()?;
+        self.flush()?;
         incremental_gc(
             &mut self.conn,
             usize::MAX,
@@ -491,17 +535,32 @@ where
         Ok(())
     }
 
+    fn maybe_checkpoint(&mut self) -> Result<()> {
+        if self.recompute_done.load(Ordering::SeqCst) {
+            self.conn
+                .pragma_update(None, "journal_size_limit", 10_000_000i64)
+                .ctx("setting journal_size_limit")?;
+            self.conn
+                .pragma_update(None, "wal_checkpoint", &"RESTART")
+                .ctx("running wal_checkpoint(RESTART)")?;
+        }
+        Ok(())
+    }
+
     /// Perform an incremental garbage collection.
     ///
     /// Will collect unpinned blocks until either the size targets are met again, or at minimum
-    /// `min_blocks` blocks are collected. Then it will continue connecting blocks until `max_duration`
+    /// `min_blocks` blocks are collected. Then it will continue collecting blocks until `max_duration`
     /// is elapsed.
     ///
     /// Note that this might significantly exceed `max_duration` for various reasons.
     ///
     /// Returns true if either size targets are met or there are no unpinned blocks left.
     pub fn incremental_gc(&mut self, min_blocks: usize, max_duration: Duration) -> Result<bool> {
+        let stats = self.get_store_stats()?;
+        let _span = tracing::debug_span!("incGC", stats = ?&stats).entered();
         self.cleanup_temp_pins()?;
+        self.maybe_checkpoint()?;
         let ret = incremental_gc(
             &mut self.conn,
             min_blocks,
@@ -509,10 +568,16 @@ where
             self.config.size_targets,
             &self.config.cache_tracker,
         )?;
-        in_txn(&mut self.conn, None, |txn| {
-            txn.execute_batch("PRAGMA incremental_vacuum")
-                .ctx("incremental vacuum")
-        })?;
+        self.maybe_checkpoint()?;
+        in_txn(
+            &mut self.conn,
+            Some(("incremental_vacuum", Duration::from_millis(500))),
+            false,
+            |txn| {
+                txn.execute_batch("PRAGMA incremental_vacuum")
+                    .ctx("incremental vacuum")
+            },
+        )?;
         Ok(ret)
     }
 }
@@ -533,16 +598,24 @@ where
     S: StoreParams,
     Ipld: References<S::Codecs>,
 {
-    delegate! {
-        /// Set or delete an alias
-        alias(name: impl AsRef<[u8]>, link: Option<&Cid>) -> Result<()>;
+    /// Set or delete an alias
+    pub fn alias<'b>(
+        &mut self,
+        name: impl Into<Cow<'b, [u8]>>,
+        link: Option<&'b Cid>,
+    ) -> Result<()> {
+        self.transaction().alias(name, link)
+    }
 
+    /// Resolves an alias to a cid
+    pub fn resolve<'b>(&mut self, name: impl Into<Cow<'b, [u8]>>) -> Result<Option<Cid>> {
+        self.transaction().resolve(name)
+    }
+
+    delegate! {
         /// Returns the aliases referencing a cid
         reverse_alias(cid: &Cid) -> Result<Option<HashSet<Vec<u8>>>>;
 
-        /// Resolves an alias to a cid
-        resolve(name: impl AsRef<[u8]>) -> Result<Option<Cid>>;
-
         /// Extend temp pin with an additional cid
         extend_temp_pin(pin: &mut TempPin, link: &Cid) -> Result<()>;
 
@@ -572,7 +645,7 @@ where
         /// Put a block
         ///
         /// This will only be completed once the transaction is successfully committed.
-        put_block(block: &Block<S>, pin: Option<&mut TempPin>) -> Result<()>;
+        put_block(block: Block<S>, pin: Option<&mut TempPin>) -> Result<()>;
 
         /// Get a block
         get_block(cid: &Cid) -> Result<Option<Vec<u8>>>;
@@ -589,7 +662,8 @@ where
     {
         let mut txn = self.transaction();
         for block in blocks {
-            txn.put_block(&block, pin.as_deref_mut())?;
+            #[allow(clippy::needless_option_as_deref)]
+            txn.put_block(block, pin.as_deref_mut())?;
         }
         txn.commit()
     }
